from io import BytesIO, StringIO
from random import random

import math
import numpy as np
import matplotlib.pyplot as plt

from src.normativeKnowledge import NormativeKnowledge
from src.situationalKnowledge import SituationalKnowledge
from src.topologicalKnowledge import TopologicalKnowledge


class Population:

    def __init__(self, number, world, default=True, situational=False, normative=False, topological=False,
                 mutation_step_size=0.1):
        self.mutation_step_size = mutation_step_size
        self.individuals_n = number
        self.world = world
        self.x = np.empty([self.individuals_n])
        self.y = np.empty([self.individuals_n])
        self.fitness = np.empty([self.individuals_n])
        for i in range(number):
            self.x[i] = (random() * world.size_x)
            self.y[i] = (random() * world.size_y)
            self.fitness[i] = (world.height(self.x[i], self.y[i]))
        self.population_fitness = []

        # cultural components
        self.situational = None
        self.normative = None
        self.topological = None
        self.default = default

        self.groups = 0
        if default:
            self.groups += 1
        if situational:
            self.groups += 1
        if normative:
            self.groups += 1
        if topological:
            self.groups += 1
        if self.groups == 0:
            self.default = True
            self.groups = 1

        if situational:
            self.situational = SituationalKnowledge()
            self.situational.weight = 1/self.groups
        if normative:
            self.normative = NormativeKnowledge()
            self.normative.weight = 1/self.groups
        if topological:
            self.topological = TopologicalKnowledge()
            self.topological.weight = 1/self.groups

        self.knowledge = np.empty([self.individuals_n])
        self.update_knowledge_belonging(1)

    def update_knowledge_belonging(self, update_percentage):
        for i in range(self.individuals_n):
            # update only percentage
            if random() < update_percentage:
                roulette = random()
                base = 0
                if self.situational:
                    if roulette < self.situational.weight:
                        self.knowledge[i] = 1
                    base += self.situational.weight
                if self.normative:
                    if base < roulette < (base + self.normative.weight):
                        self.knowledge[i] = 2
                    base += self.normative.weight
                if self.topological:
                    if base < roulette < (base + self.topological.weight):
                        self.knowledge[i] = 3
                    base += self.topological.weight
                if base < roulette:
                    self.knowledge[i] = 0

    def update_knowledge_weights(self):
        base_weight = 0.05
        count = np.zeros([4])
        sum = np.zeros([4])
        for i in range(self.individuals_n):
            count[int(self.knowledge[i])] += 1
            sum[int(self.knowledge[i])] += self.fitness[i]
        total_sum = np.sum(sum) / (1 - self.groups * base_weight)
        if self.situational:
            self.situational.weight = base_weight + sum[1]/total_sum
            print("sit w: " + str(self.situational.weight))
        if self.normative:
            self.normative.weight = base_weight + sum[2]/total_sum
            print("norm w: " + str(self.normative.weight))
        if self.topological:
            self.topological.weight = base_weight + sum[3]/total_sum
            print("top w: " + str(self.topological.weight))

    def new_generation(self):
        self.evolve()
        if self.situational:
            self.situational.accept(self)
        if self.normative:
            self.normative.accept(self)
        if self.topological:
            self.topological.accept(self)

        if self.situational:
            self.situational.influence(self)
        if self.normative:
            self.normative.influence(self)
        if self.topological:
            self.topological.influence(self)
        self.population_fitness.append(np.sum(self.fitness) / self.individuals_n)

    def evolve(self):
        """Generates the next generation of individuals"""
        child_x = np.empty([self.individuals_n])
        child_y = np.empty([self.individuals_n])
        child_fitness = np.empty([self.individuals_n])
        for i in range(self.individuals_n):
            step = random() * self.mutation_step_size
            direction = random() * math.pi * 2
            child_x[i] = (math.cos(direction) * step) + self.x[i]
            child_y[i] = (math.sin(direction) * step) + self.y[i]
            child_fitness[i] = (self.world.height(child_x[i], child_y[i]))

        pool_x = np.concatenate((self.x, child_x))
        pool_y = np.concatenate((self.y, child_y))
        pool = np.concatenate((self.fitness, child_fitness))
        pool_relative = self.relative_fitness(pool, int(0.1 * self.individuals_n))
        selection = self.roulette_wheel_selection(pool_relative, self.individuals_n)

        self.x = pool_x[selection]
        self.y = pool_y[selection]
        self.fitness = pool[selection]
        self.knowledge = np.concatenate((self.knowledge, self.knowledge))[selection]

        self.update_knowledge_weights()
        self.update_knowledge_belonging(0.1)

    @staticmethod
    def relative_fitness(fitness_array, compare_against):
        """Returns a numpy array with the relative fitness values.

        fitness_array: numpy array with the absolute fitness
        compare_against: number of individuals that are used for comparison
        """
        result = np.empty(fitness_array.shape)
        for i in range(len(fitness_array)):
            rel_fitness = 0
            competitors = np.random.choice(fitness_array, compare_against)
            for c in competitors:
                if fitness_array[i] > c:
                    rel_fitness += 1
            result[i] = rel_fitness
        return result

    @staticmethod
    def roulette_wheel_selection(pool, amount_to_select):
        """Selects amount_to_select entries from a numpy array and returns a numpy array with the positions
        of the selected entries.

        pool: numpy array with fitness values
        """
        result = np.empty(amount_to_select)
        probability = np.empty(len(pool))
        fit_sum = np.sum(pool)
        for i in range(len(pool)):
            probability[i] = pool[i]/fit_sum

        for x in range(amount_to_select):
            roulette = random()
            r_sum = probability[0]
            choice = 0
            while r_sum <= roulette:
                choice += 1
                r_sum += probability[choice]
            result[x] = choice

        return np.int8(result)

    def plot(self, plot_vs_return=False):

        sit_x = self.x[self.knowledge == 1]
        sit_y = self.y[self.knowledge == 1]
        norm_x = self.x[self.knowledge == 2]
        norm_y = self.y[self.knowledge == 2]
        top_x = self.x[self.knowledge == 3]
        top_y = self.y[self.knowledge == 3]
        def_x = self.x[self.knowledge == 0]
        def_y = self.y[self.knowledge == 0]
        print("numbers sit:" + str(len(sit_x)) + " norm: " + str(len(norm_x)) + " top: " + str(len(top_x)) + " def: " + str(len(def_x)))
        print("total " + str(len(sit_x)+len(norm_x)+len(top_x)+len(def_x)))
        fig, ax = plt.subplots()
        ax.set_xlim(0, self.world.size_x)
        ax.set_ylim(0, self.world.size_y)
        ax.scatter(norm_x, norm_y, marker='o')
        ax.scatter(def_x, def_y, marker='v')
        ax.scatter(sit_x, sit_y, marker='+')
        ax.scatter(top_x, top_y, marker='x')
        ax.scatter(self.situational.best_x, self.situational.best_y, marker="*", c='black')
        plt.xlabel("x")
        plt.ylabel("y")

        if plot_vs_return:
            plt.show()
            plt.clf()
            plt.cla()
            plt.close()
        else:
            img = BytesIO()
            plt.savefig(img, format='png')
            plt.clf()
            plt.cla()
            plt.close(fig)
            img.seek(0)
            return img

    def plot_fitness(self, plot_vs_return=False):
        fit_fig, fit_ax = plt.subplots()
        fit_ax.plot(range(len(self.population_fitness)), self.population_fitness)
        plt.xlabel("iteration")
        plt.ylabel("average fitness")

        if plot_vs_return:
            plt.show()
            plt.close(fit_fig)
        else:
            img = BytesIO()
            fit_fig.savefig(img, format='png')
            plt.close(fit_fig)
            img.seek(0)
            return img
